#ifndef COMTYP_H
#define COMTYP_H
#include "const.h"

typedef struct
{
    long long time;          // time (s) expressed by standard time_t
    double sec;              // fraction of second under 1 s 
}gtime_t;

typedef struct
{
    int sys;                 // navigation system 
    int vs;                  // valid satellite flag single 
    double azel[2];          // azimuth/elevation angles {az,el} (rad) 
    double resp[NFREQ];      // residuals of pseudorange (m) 
    double resc[NFREQ];      // residuals of carrier-phase (m) 
    double icbias[NFREQ];    // glonass IC bias (cycles) 
    int vsat[NFREQ];         // valid satellite flag 
    int snr[NFREQ];          // signal strength (0.25 dBHz) 
    int lock[NFREQ];         // lock counter of phase 
}ssat_t;

typedef struct 
{
    gtime_t time;            // receiver sampling time (GPST) 
    int sat,rcv;             // satellite/receiver number 
    int SNR[NFREQ+NEXOBS];   // signal strength (0.25 dBHz) 
    int LLI[NFREQ+NEXOBS];   // loss of lock indicator 
    int code[NFREQ+NEXOBS];  // code indicator (CODE_???) 
    int qualL[NFREQ+NEXOBS]; // quality of carrier phase measurement 
    int qualP[NFREQ+NEXOBS]; // quality of pseudorange measurement 
    double L[NFREQ+NEXOBS];  // observation data carrier-phase (cycle) 
    double P[NFREQ+NEXOBS];  // observation data pseudorange (m) 
    double D[NFREQ+NEXOBS];  // observation data doppler frequency (Hz) 
}obsd_t;

typedef struct 
{
    int n,nmax;              // number of obervation data/allocated 
    double tint;             // time interval
    obsd_t* data;          // observation data records 
}obs_t;

typedef struct 
{
    int sat;                 // satellite number 
    int iode,iodc;           // IODE,IODC 
    int sva;                 // SV accuracy (URA index) 
    int svh;                 // SV health (0:ok) 
    int week;                // GPS/QZS: gps week, GAL: galileo week 
    int code;                // GPS/QZS: code on L2, GAL/CMP: data sources 
    int flag;                // GPS/QZS: L2 P data flag, CMP: nav type 
    gtime_t toe,toc,ttr;     // Toe,Toc,T_trans 
                             // SV orbit parameters 
    double A,e,i0,OMG0,omg,M0,deln,OMGd,idot;
    double crc,crs,cuc,cus,cic,cis;
    double toes;             // Toe (s) in week 
    double fit;              // fit interval (h) 
    double f0,f1,f2;         // SV clock parameters (af0,af1,af2) 
    double tgd[4];           // group delay parameters 
                             // GPS/QZS:tgd(0)=TGD 
                             // GAL    :tgd(0)=BGD E5a/E1,tgd(1)=BGD E5b/E1 
                             // CMP    :tgd(0)=BGD1,tgd(1)=BGD2 
    double Adot,ndot;        // Adot,ndot for CNAV
}eph_t;

typedef struct 
{
    int sat;                 // satellite number 
    int iode;                // IODE (0-6 bit of tb field) 
    int frq;                 // satellite frequency number 
    int svh,sva,age;         // satellite health, accuracy, age of operation 
    gtime_t toe;             // epoch of epherides (gpst) 
    gtime_t tof;             // message frame time (gpst) 
    double pos[3];           // satellite position (ecef) (m) 
    double vel[3];           // satellite velocity (ecef) (m/s) 
    double acc[3];           // satellite acceleration (ecef) (m/s^2) 
    double taun,gamn;        // SV clock bias (s)/relative freq bias 
    double dtaun;            // delay between L1 and L2 (s) 
}geph_t;

typedef struct
{                            /* SBAS ephemeris type */
    int sat;                 /* satellite number */
    gtime_t t0;              /* reference epoch time (GPST) */
    gtime_t tof;             /* time of message frame (GPST) */
    int sva;                 /* SV accuracy (URA index) */
    int svh;                 /* SV health (0:ok) */
    double pos[3];           /* satellite position (m) (ecef) */
    double vel[3];           /* satellite velocity (m/s) (ecef) */
    double acc[3];           /* satellite acceleration (m/s^2) (ecef) */
    double af0, af1;         /* satellite clock-offset/drift (s,s/s) */
}seph_t;

typedef struct 
{
    int sat;                 // satellite number 
    int svh;                 // sv health (0:ok) 
    int svconf;              // as and sv config 
    int week;                // GPS/QZS: gps week, GAL: galileo week 
    gtime_t toa;             // Toa 
                             // SV orbit parameters 
    double A,e,i0,OMG0,omg,M0,OMGd;
    double toas;             // Toa (s) in week 
    double f0,f1;            // SV clock parameters (af0,af1)
}alm_t;

typedef struct 
{
    int n,nmax;              // number of broadcast ephemeris 
    int ng,ngmax;            // number of glonass ephemeris 
    int ns,nsmax;            // number of sbas ephemeris 
    int ne,nemax;            // number of precise ephemeris 
    int nc,ncmax;            // number of precise clock 
    int na,namax;            // number of almanac data 
    int nt,ntmax;            // number of tec grid data 
    int nf,nfmax;            // number of satellite fcb data 
    eph_t* eph;              // GPS/QZS/GAL ephemeris 
    geph_t* geph;            // GLONASS ephemeris 
    seph_t *seph;            /* SBAS ephemeris */
    alm_t* alm;              // almanac data 
    double utc_gps[4];       // GPS delta-UTC parameters {A0,A1,T,W} 
    double utc_glo[4];       // GLONASS UTC GPS time parameters 
    double utc_gal[4];       // Galileo UTC GPS time parameters 
    double utc_qzs[4];       // QZS UTC GPS time parameters 
    double utc_cmp[4];       // BeiDou UTC parameters 
    double utc_irn[4];       // IRNSS UTC parameters 
    double utc_sbs[4];       // SBAS UTC parameters 
    double ion_gps[8];       // GPS iono model parameters {a0,a1,a2,a3,b0,b1,b2,b3} 
    double ion_gal[4];       // Galileo iono model parameters {ai0,ai1,ai2,0} 
    double ion_qzs[8];       // QZSS iono model parameters {a0,a1,a2,a3,b0,b1,b2,b3} 
    double ion_cmp[8];       // BeiDou iono model parameters {a0,a1,a2,a3,b0,b1,b2,b3} 
    double ion_irn[8];       // IRNSS iono model parameters {a0,a1,a2,a3,b0,b1,b2,b3} 
    int leaps;               // leap seconds (s) 
    double lam[MAXSAT][NFREQ];// carrier wave lengths (m) 
    double cbias[MAXSAT][3]; // satellite dcb (0:p1-p2,1:p1-c1,2:p2-c2) (m) 
    double rbias[MAXRCV][2][3];// receiver dcb (0:p1-p2,1:p1-c1,2:p2-c2) (m) 
    double wlbias[MAXSAT];   // wide-lane bias (cycle) 
    double glo_cpbias[4];    // glonass code-phase bias {1C,1P,2C,2P} (m) 
    char glo_fcn[MAXPRNGLO+1];// glonass frequency channel number + 8 
}nav_t;

typedef struct 
{
    char name[MAXANT];       // marker name 
    char marker[MAXANT];     // marker number 
    char antdes[MAXANT];     // antenna descriptor 
    char antsno[MAXANT];     // antenna serial number 
    char rectype[MAXANT];    // receiver type descriptor 
    char recver[MAXANT];     // receiver firmware version 
    char recsno[MAXANT];     // receiver serial number 
    int antsetup;            // antenna setup id 
    int itrf;                // ITRF realization year 
    int deltype;             // antenna delta type (0:enu,1:xyz) 
    double pos[3];           // station position (ecef) (m) 
    double del[3];           // antenna position delta (e/n/u or x/y/z) (m) 
    double hgt;              // antenna height (m) 
}sta_t;

typedef struct 
{
    int n;                   //number of index 
    int frqIdx[MAXOBSTYPE];  //signal frequency (1:L1,2:L2,...) 
    int pos[MAXOBSTYPE];     //signal index in obs data (-1:no) 
    int pri[MAXOBSTYPE];     //signal priority (15-0) 
    int type[MAXOBSTYPE];    //type (0:C,1:L,2:D,3:S) 
    int code[MAXOBSTYPE];    //obs code (CODE_L??) 
    double shift[MAXOBSTYPE];//phase shift (cycle) 
}sigind_t;


#endif