#include <stdlib.h>
#include "decodenav.h"
#include "str.h"
#include "timeconvert.h"

/* URA value (m) to URA index ------------------------------------------------*/
static int uraindex(double value)
{
    int i;
    for (i = 0; i < 15; i++){
        if (ura_eph[i] >= value) break;
    }
    return i;
}

/* Galileo SISA value (m) to SISA index --------------------------------------*/
static int sisa_index(double value)
{
    /*
     * kudos to https://core.ac.uk/download/pdf/328854682.pdf for this ...
     * Signal-in-Space Accuracy : SISA flag is a prediction at 1-sigma standard deviation of the quality of
     * the transmitted signal.The flag can take values from 0 to 255. The transmitted standard index is 107 that
     * corresponds to a SISA value of 3.12m.If the prediction is not available, the transmitted index is 255 and
     * corresponds to No Accurate Prediction Available(NAPA).NAPA is an indicator of a potential anomalous
     * signal-in-space[6].Please notice that SISA flag refers to the dual-frequency signal combinations.
     */
    if (value < 0.0 || value > 6.0)
        return 255; /* unknown or NAPA */
    else if (value <= 0.5)
        return (int)(value / 0.01);
    else if (value <= 1.0)
        return (int)((value - 0.5) / 0.02) + 50;
    else if (value <= 2.0)
        return (int)((value - 1.0) / 0.04) + 75;
    return (int)((value - 2.0) / 0.16) + 100;
}

void decode_navh(char *buff, nav_t *nav)
{
    int i,j;
    char* label = buff + 60;
    if (strstr(label, "ION ALPHA"))
    { /* opt ver.2 */
        if (nav)
        {
            for (i = 0, j = 2; i < 4; i++, j += 12)
            {
                nav->ion_gps[i] = str2num(buff, j, 12);
            }  
        }
    }
    else if (strstr(label, "ION BETA"))
    { /* opt ver.2 */
        if (nav)
        {
            for (i = 0, j = 2; i < 4; i++, j += 12)
            {    
                nav->ion_gps[i + 4] = str2num(buff, j, 12);
            }
        }
    }
    else if (strstr(label, "DELTA-UTC: A0,A1,T,W"))
    { /* opt ver.2 */
        if (nav)
        {
            for (i = 0, j = 3; i < 2; i++, j += 19)
            {    
                nav->utc_gps[i] = str2num(buff, j, 19);
            }
            for (; i < 4; i++, j += 9)
            {
                nav->utc_gps[i] = str2num(buff, j, 9);
            }
        }
    }
    else if (strstr(label, "IONOSPHERIC CORR"))
    { /* opt ver.3 */
        if (nav)
        {
            if (!strncmp(buff, "GPSA", 4))
            {
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_gps[i] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "GPSB", 4))
            {
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_gps[i + 4] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "GAL", 3))
            {
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_gal[i] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "QZSA", 4))
            { /* v.3.02 */
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_qzs[i] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "QZSB", 4))
            { /* v.3.02 */
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_qzs[i + 4] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "BDSA", 4))
            { /* v.3.02 */
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_cmp[i] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "BDSB", 4))
            { /* v.3.02 */
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_cmp[i + 4] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "IRNA", 4))
            { /* v.3.03 */
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_irn[i] = str2num(buff, j, 12);
                }
            }
            else if (!strncmp(buff, "IRNB", 4))
            { /* v.3.03 */
                for (i = 0, j = 5; i < 4; i++, j += 12)
                {    
                    nav->ion_irn[i + 4] = str2num(buff, j, 12);
                }
            }
        }
    }
    else if (strstr(label, "TIME SYSTEM CORR"))
    { /* opt ver.3 */
        if (nav)
        {
            if (!strncmp(buff, "GPUT", 4))
            {
                nav->utc_gps[0] = str2num(buff, 5, 17);
                nav->utc_gps[1] = str2num(buff, 22, 16);
                nav->utc_gps[2] = str2num(buff, 38, 7);
                nav->utc_gps[3] = str2num(buff, 45, 5);
            }
            else if (!strncmp(buff, "GLUT", 4))
            {
                nav->utc_glo[0] = -str2num(buff, 5, 17); /* tau_C */
            }
            else if (!strncmp(buff, "GLGP", 4))
            {
                nav->utc_glo[1] = str2num(buff, 5, 17); /* tau_GPS */
            }
            else if (!strncmp(buff, "GAUT", 4))
            { /* v.3.02 */
                nav->utc_gal[0] = str2num(buff, 5, 17);
                nav->utc_gal[1] = str2num(buff, 22, 16);
                nav->utc_gal[2] = str2num(buff, 38, 7);
                nav->utc_gal[3] = str2num(buff, 45, 5);
            }
            else if (!strncmp(buff, "QZUT", 4))
            { /* v.3.02 */
                nav->utc_qzs[0] = str2num(buff, 5, 17);
                nav->utc_qzs[1] = str2num(buff, 22, 16);
                nav->utc_qzs[2] = str2num(buff, 38, 7);
                nav->utc_qzs[3] = str2num(buff, 45, 5);
            }
            else if (!strncmp(buff, "BDUT", 4))
            { /* v.3.02 */
                nav->utc_cmp[0] = str2num(buff, 5, 17);
                nav->utc_cmp[1] = str2num(buff, 22, 16);
                nav->utc_cmp[2] = str2num(buff, 38, 7);
                nav->utc_cmp[3] = str2num(buff, 45, 5);
            }
            else if (!strncmp(buff, "SBUT", 4))
            { /* v.3.02 */
                nav->utc_sbs[0] = str2num(buff, 5, 17);
                nav->utc_sbs[1] = str2num(buff, 22, 16);
                nav->utc_sbs[2] = str2num(buff, 38, 7);
                nav->utc_sbs[3] = str2num(buff, 45, 5);
            }
            else if (!strncmp(buff, "IRUT", 4))
            { /* v.3.03 */
                nav->utc_irn[0] = str2num(buff, 5, 17);
                nav->utc_irn[1] = str2num(buff, 22, 16);
                nav->utc_irn[2] = str2num(buff, 38, 7);
                nav->utc_irn[3] = str2num(buff, 45, 5);
            }
        }
    }
    else if (strstr(label, "LEAP SECONDS"))
    { /* opt */
        if (nav)
        {
            nav->leaps = (int)str2num(buff, 0, 6);
        }
    }
    return;
}

void decode_gnavh(char *buff, nav_t *nav)
{
    char *label = buff + 60;

    if (strstr(label, "CORR TO SYSTEM TIME"));
    else if (strstr(label, "LEAP SECONDS"))
    {
        if (nav)
        {
            nav->leaps = (int)str2num(buff, 0, 6);
        }
    }
}

int decode_geph(double ver, int sat, gtime_t toc, double *data, geph_t *geph)
{
    geph_t geph0 = {0};
    gtime_t tof;
    double tow, tod;
    int week, dow;

    printf("decode_geph: ver=%.2f sat=%2d\n", ver, sat);

    if (satsys(sat, NULL) != SYS_GLO)
    {
        printf("glonass ephemeris error: invalid satellite sat=%2d\n", sat);
        return 0;
    }
    *geph = geph0;

    geph->sat = sat;

    /* Toc rounded by 15 min in utc */
    tow = time2gpst(toc, &week);
    toc = gpst2time(week, floor((tow + 450.0) / 900.0) * 900);
    dow = (int)floor(tow / 86400.0);

    /* time of frame in UTC */
    tod = ver <= 2.99 ? data[2] : fmod(data[2], 86400.0); /* Tod (v.2), Tow (v.3) in UTC */
    tof = gpst2time(week, tod + dow * 86400.0);
    tof = adjday(tof, toc);

    geph->toe = utc2gpst(toc); /* Toc (GPST) */
    geph->tof = utc2gpst(tof); /* Tof (GPST) */

    /* IODE = Tb (7bit), Tb =index of UTC+3H within current day */
    geph->iode = (int)(fmod(tow + 10800.0, 86400.0) / 900.0 + 0.5);

    geph->taun = -data[0]; /* -taun */
    geph->gamn = data[1];  /* +gamman */

    geph->pos[0] = data[3] * 1E3;
    geph->pos[1] = data[7] * 1E3;
    geph->pos[2] = data[11] * 1E3;
    geph->vel[0] = data[4] * 1E3;
    geph->vel[1] = data[8] * 1E3;
    geph->vel[2] = data[12] * 1E3;
    geph->acc[0] = data[5] * 1E3;
    geph->acc[1] = data[9] * 1E3;
    geph->acc[2] = data[13] * 1E3;

    geph->svh = (int)data[6];
    geph->frq = (int)data[10];

    geph->age = (int)data[14];

    /* some receiver output >128 for minus frequency number */
    if (geph->frq > 128)
        geph->frq -= 256;

    if (geph->frq < MINFREQ_GLO || MAXFREQ_GLO < geph->frq)
    {
        printf("rinex gnav invalid freq: sat=%2d fn=%d\n", sat, geph->frq);
    }
    return 1;
}

int decode_seph(double ver, int sat, gtime_t toc, double *data,seph_t *seph)
{
    seph_t seph0 = {0};
    int week;

    printf("decode_seph: ver=%.2f sat=%2d\n", ver, sat);

    if (satsys(sat, NULL) != SYS_SBS)
    {
        printf("geo ephemeris error: invalid satellite sat=%2d\n", sat);
        return 0;
    }
    *seph = seph0;

    seph->sat = sat;
    seph->t0 = toc;

    time2gpst(toc, &week);
    seph->tof = adjweek(gpst2time(week, data[2]), toc);

    seph->af0 = data[0];
    seph->af1 = data[1];

    seph->pos[0] = data[3] * 1E3;
    seph->pos[1] = data[7] * 1E3;
    seph->pos[2] = data[11] * 1E3;
    seph->vel[0] = data[4] * 1E3;
    seph->vel[1] = data[8] * 1E3;
    seph->vel[2] = data[12] * 1E3;
    seph->acc[0] = data[5] * 1E3;
    seph->acc[1] = data[9] * 1E3;
    seph->acc[2] = data[13] * 1E3;

    seph->svh = (int)data[6];
    seph->sva = uraindex(data[10]);

    return 1;
}

int decode_eph(double ver, int sat, gtime_t toc, const double *data,
                      eph_t *eph)
{
    eph_t eph0 = {0};
    int sys;

    printf("decode_eph: ver=%.2f sat=%2d\n", ver, sat);

    sys = satsys(sat, NULL);

    if (!(sys & (SYS_GPS | SYS_GAL | SYS_QZS | SYS_CMP | SYS_IRN)))
    {
        printf("ephemeris error: invalid satellite sat=%2d\n", sat);
        return 0;
    }
    *eph = eph0;

    eph->sat = sat;
    eph->toc = toc;

    eph->f0 = data[0];
    eph->f1 = data[1];
    eph->f2 = data[2];

    eph->A = data[10]*data[10];
    eph->e = data[8];
    eph->i0 = data[15];
    eph->OMG0 = data[13];
    eph->omg = data[17];
    eph->M0 = data[6];
    eph->deln = data[5];
    eph->OMGd = data[18];
    eph->idot = data[19];
    eph->crc = data[16];
    eph->crs = data[4];
    eph->cuc = data[7];
    eph->cus = data[9];
    eph->cic = data[12];
    eph->cis = data[14];

    if (sys == SYS_GPS || sys == SYS_QZS)
    {
        eph->iode = (int)data[3];  /* IODE */
        eph->iodc = (int)data[26]; /* IODC */
        eph->toes = data[11];      /* Toe (s) in GPS week */
        eph->week = (int)data[21]; /* GPS week */
        eph->toe = adjweek(gpst2time(eph->week, data[11]), toc);
        eph->ttr = adjweek(gpst2time(eph->week, data[27]), toc);

        eph->code = (int)data[20];     /* GPS: codes on L2 ch */
        eph->svh = (int)data[24];      /* SV health */
        eph->sva = uraindex(data[23]); /* URA index (m->index) */
        eph->flag = (int)data[22];     /* GPS: L2 P data flag */

        eph->tgd[0] = data[25]; /* TGD */
        if (sys == SYS_GPS)
        {
            eph->fit = data[28]; /* fit interval (h) */
        }
        else
        {
            eph->fit = data[28] == 0.0 ? 1.0 : 2.0; /* fit interval (0:1h,1:>2h) */
        }
    }
    else if (sys == SYS_GAL)
    {                              /* GAL ver.3 */
        eph->iode = (int)data[3];  /* IODnav */
        eph->toes = data[11];      /* Toe (s) in Galileo week */
        eph->week = (int)data[21]; /* Galileo week = GPS week */
        eph->toe = adjweek(gpst2time(eph->week, data[11]), toc);
        eph->ttr = adjweek(gpst2time(eph->week, data[27]), toc);

        eph->code = (int)data[20];       /* data sources */
                                         /* bit 0 set: I/NAV E1-B */
                                         /* bit 1 set: F/NAV E5a-I */
                                         /* bit 2 set: F/NAV E5b-I */
                                         /* bit 8 set: af0-af2 toc are for E5a.E1 */
                                         /* bit 9 set: af0-af2 toc are for E5b.E1 */
        eph->svh = (int)data[24];        /* sv health */
                                         /* bit     0: E1B DVS */
                                         /* bit   1-2: E1B HS */
                                         /* bit     3: E5a DVS */
                                         /* bit   4-5: E5a HS */
                                         /* bit     6: E5b DVS */
                                         /* bit   7-8: E5b HS */
        eph->sva = sisa_index(data[23]); /* sisa (m->index) */

        eph->tgd[0] = data[25]; /* BGD E5a/E1 */
        eph->tgd[1] = data[26]; /* BGD E5b/E1 */
    }
    else if (sys == SYS_CMP)
    {                                                       /* BeiDou v.3.02 */
        eph->toc = bdt2gpst(eph->toc);                      /* bdt -> gpst */
        eph->iode = (int)data[3];                           /* AODE */
        eph->iodc = (int)data[28];                          /* AODC */
        eph->toes = data[11];                               /* Toe (s) in BDT week */
        eph->week = (int)data[21];                          /* bdt week */
        eph->toe = bdt2gpst(bdt2time(eph->week, data[11])); /* BDT -> GPST */
        eph->ttr = bdt2gpst(bdt2time(eph->week, data[27])); /* BDT -> GPST */
        eph->toe = adjweek(eph->toe, toc);
        eph->ttr = adjweek(eph->ttr, toc);

        eph->svh = (int)data[24];      /* satH1 */
        eph->sva = uraindex(data[23]); /* URA index (m->index) */

        eph->tgd[0] = data[25]; /* TGD1 B1/B3 */
        eph->tgd[1] = data[26]; /* TGD2 B2/B3 */
    }
    else if (sys == SYS_IRN)
    {                              /* IRNSS v.3.03 */
        eph->iode = (int)data[3];  /* IODEC */
        eph->toes = data[11];      /* Toe (s) in IRNSS week */
        eph->week = (int)data[21]; /* IRNSS week */
        eph->toe = adjweek(gpst2time(eph->week, data[11]), toc);
        eph->ttr = adjweek(gpst2time(eph->week, data[27]), toc);
        eph->svh = (int)data[24];      /* SV health */
        eph->sva = uraindex(data[23]); /* URA index (m->index) */
        eph->tgd[0] = data[25];        /* TGD */
    }
    if (eph->iode < 0 || 1023 < eph->iode)
    {
        printf("rinex nav invalid: sat=%2d iode=%d\n", sat, eph->iode);
    }
    if (eph->iodc < 0 || 1023 < eph->iodc)
    {
        printf("rinex nav invalid: sat=%2d iodc=%d\n", sat, eph->iodc);
    }
    return 1;
}