#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "readobsnav.h"
#include "init_val.h"
#include "decodeobs.h"
#include "decodenav.h"
#include "str.h"
#include "timeconvert.h"

/* add ephemeris to navigation data ------------------------------------------*/
static int add_eph(nav_t *nav, const eph_t *eph)
{
    eph_t *nav_eph;

    if (nav->nmax <= nav->n)
    {
        nav->nmax += 1024;
        if (!(nav_eph = (eph_t *)realloc(nav->eph, sizeof(eph_t) * nav->nmax)))
        {
            printf("decode_eph malloc error: n=%d\n", nav->nmax);
            free(nav->eph);
            nav->eph = NULL;
            nav->n = nav->nmax = 0;
            return 0;
        }
        nav->eph = nav_eph;
    }
    nav->eph[nav->n++] = *eph;
    return 1;
}

/* compare observation data -------------------------------------------------*/
static int cmpobs(const void *p1, const void *p2)
{
    obsd_t *q1 = (obsd_t *)p1, *q2 = (obsd_t *)p2;
    double tt = timediff(q1->time, q2->time);
    if (fabs(tt) > DTTOL)
        return tt < 0 ? -1 : 1;
    if (q1->rcv != q2->rcv)
        return (int)q1->rcv - (int)q2->rcv;
    return (int)q1->sat - (int)q2->sat;
}

/* sort and unique observation data --------------------------------------------
 * sort and unique observation data by time, rcv, sat
 * args   : obs_t *obs    IO     observation data
 * return : number of epochs
 *-----------------------------------------------------------------------------*/
int sortobs(obs_t *obs)
{
    int i, j, n;

    printf("sortobs: nobs=%d\n", obs->n);

    if (obs->n <= 0)
        return 0;

    qsort(obs->data, obs->n, sizeof(obsd_t), cmpobs);

    /* delete duplicated data */
    for (i = j = 0; i < obs->n; i++)
    {
        if (obs->data[i].sat != obs->data[j].sat ||
            obs->data[i].rcv != obs->data[j].rcv ||
            timediff(obs->data[i].time, obs->data[j].time) != 0.0)
        {
            obs->data[++j] = obs->data[i];
        }
    }
    obs->n = j + 1;

    for (i = n = 0; i < obs->n; i = j, n++)
    {
        for (j = i + 1; j < obs->n; j++)
        {
            if (timediff(obs->data[j].time, obs->data[i].time) > DTTOL)
                break;
        }
    }
    return n;
}

/* compare ephemeris ---------------------------------------------------------*/
static int cmpeph(const void *p1, const void *p2)
{
    eph_t *q1 = (eph_t *)p1, *q2 = (eph_t *)p2;
    return q1->ttr.time != q2->ttr.time ? (int)(q1->ttr.time - q2->ttr.time) : (q1->toe.time != q2->toe.time ? (int)(q1->toe.time - q2->toe.time) : q1->sat - q2->sat);
}
/* sort and unique ephemeris -------------------------------------------------*/
static void uniqeph(nav_t *nav)
{
    eph_t *nav_eph;
    int i, j;

    printf("uniqeph: n=%d\n", nav->n);

    if (nav->n <= 0)
        return;

    qsort(nav->eph, nav->n, sizeof(eph_t), cmpeph);

    for (i = 1, j = 0; i < nav->n; i++)
    {
        if (nav->eph[i].sat != nav->eph[j].sat ||
            nav->eph[i].iode != nav->eph[j].iode)
        {
            nav->eph[++j] = nav->eph[i];
        }
    }
    nav->n = j + 1;

    if (!(nav_eph = (eph_t *)realloc(nav->eph, sizeof(eph_t) * nav->n)))
    {
        printf("uniqeph malloc error n=%d\n", nav->n);
        free(nav->eph);
        nav->eph = NULL;
        nav->n = nav->nmax = 0;
        return;
    }
    nav->eph = nav_eph;
    nav->nmax = nav->n;

    printf("uniqeph: n=%d\n", nav->n);
}
/* compare glonass ephemeris -------------------------------------------------*/
static int cmpgeph(const void *p1, const void *p2)
{
    geph_t *q1 = (geph_t *)p1, *q2 = (geph_t *)p2;
    return q1->tof.time != q2->tof.time ? (int)(q1->tof.time - q2->tof.time) : (q1->toe.time != q2->toe.time ? (int)(q1->toe.time - q2->toe.time) : q1->sat - q2->sat);
}
/* sort and unique glonass ephemeris -----------------------------------------*/
static void uniqgeph(nav_t *nav)
{
    geph_t *nav_geph;
    int i, j;

    printf("uniqgeph: ng=%d\n", nav->ng);

    if (nav->ng <= 0)
        return;

    qsort(nav->geph, nav->ng, sizeof(geph_t), cmpgeph);

    for (i = j = 0; i < nav->ng; i++)
    {
        if (nav->geph[i].sat != nav->geph[j].sat ||
            nav->geph[i].toe.time != nav->geph[j].toe.time ||
            nav->geph[i].svh != nav->geph[j].svh)
        {
            nav->geph[++j] = nav->geph[i];
        }
    }
    nav->ng = j + 1;

    if (!(nav_geph = (geph_t *)realloc(nav->geph, sizeof(geph_t) * nav->ng)))
    {
        printf("uniqgeph malloc error ng=%d\n", nav->ng);
        free(nav->geph);
        nav->geph = NULL;
        nav->ng = nav->ngmax = 0;
        return;
    }
    nav->geph = nav_geph;
    nav->ngmax = nav->ng;

    printf("uniqgeph: ng=%d\n", nav->ng);
}
/* compare sbas ephemeris ----------------------------------------------------*/
static int cmpseph(const void *p1, const void *p2)
{
    seph_t *q1 = (seph_t *)p1, *q2 = (seph_t *)p2;
    return q1->tof.time != q2->tof.time ? (int)(q1->tof.time - q2->tof.time) : (q1->t0.time != q2->t0.time ? (int)(q1->t0.time - q2->t0.time) : q1->sat - q2->sat);
}
/* sort and unique sbas ephemeris --------------------------------------------*/
static void uniqseph(nav_t *nav)
{
    seph_t *nav_seph;
    int i, j;

    printf("uniqseph: ns=%d\n", nav->ns);

    if (nav->ns <= 0)
        return;

    qsort(nav->seph, nav->ns, sizeof(seph_t), cmpseph);

    for (i = j = 0; i < nav->ns; i++)
    {
        if (nav->seph[i].sat != nav->seph[j].sat ||
            nav->seph[i].t0.time != nav->seph[j].t0.time)
        {
            nav->seph[++j] = nav->seph[i];
        }
    }
    nav->ns = j + 1;

    if (!(nav_seph = (seph_t *)realloc(nav->seph, sizeof(seph_t) * nav->ns)))
    {
        printf("uniqseph malloc error ns=%d\n", nav->ns);
        free(nav->seph);
        nav->seph = NULL;
        nav->ns = nav->nsmax = 0;
        return;
    }
    nav->seph = nav_seph;
    nav->nsmax = nav->ns;

    printf("uniqseph: ns=%d\n", nav->ns);
}

/* unique ephemerides ----------------------------------------------------------
 * unique ephemerides in navigation data and update carrier wave length
 * args   : nav_t *nav    IO     navigation data
 * return : number of epochs
 *-----------------------------------------------------------------------------*/
void uniqnav(nav_t *nav)
{
    printf("uniqnav: neph=%d ngeph=%d nseph=%d\n", nav->n, nav->ng, nav->ns);

    /* unique ephemeris */
    uniqeph(nav);
    uniqgeph(nav);
    uniqseph(nav);
}

static int add_geph(nav_t *nav, const geph_t *geph)
{
    geph_t *nav_geph;

    if (nav->ngmax <= nav->ng)
    {
        nav->ngmax += 1024;
        if (!(nav_geph = (geph_t *)realloc(nav->geph, sizeof(geph_t) * nav->ngmax)))
        {
            printf("decode_geph malloc error: n=%d\n", nav->ngmax);
            free(nav->geph);
            nav->geph = NULL;
            nav->ng = nav->ngmax = 0;
            return 0;
        }
        nav->geph = nav_geph;
    }
    nav->geph[nav->ng++] = *geph;
    return 1;
}
static int add_seph(nav_t *nav, const seph_t *seph)
{
    seph_t *nav_seph;

    if (nav->nsmax <= nav->ns)
    {
        nav->nsmax += 1024;
        if (!(nav_seph = (seph_t *)realloc(nav->seph, sizeof(seph_t) * nav->nsmax)))
        {
            printf("decode_seph malloc error: n=%d\n", nav->nsmax);
            free(nav->seph);
            nav->seph = NULL;
            nav->ns = nav->nsmax = 0;
            return 0;
        }
        nav->seph = nav_seph;
    }
    nav->seph[nav->ns++] = *seph;
    return 1;
}

/* set system mask -----------------------------------------------------------*/
static int set_sysmask(const char *opt)
{
    const char *p;
    int mask = SYS_NONE;

    if (!(p = strstr(opt, "-SYS=")))
        return SYS_ALL;

    for (p += 5; *p && *p != ' '; p++)
    {
        switch (*p)
        {
        case 'G':
            mask |= SYS_GPS;
            break;
        case 'R':
            mask |= SYS_GLO;
            break;
        case 'E':
            mask |= SYS_GAL;
            break;
        case 'J':
            mask |= SYS_QZS;
            break;
        case 'C':
            mask |= SYS_CMP;
            break;
        case 'I':
            mask |= SYS_IRN;
            break;
        case 'S':
            mask |= SYS_SBS;
            break;
        }
    }
    return mask;
}

/* obs type string to obs code -------------------------------------------------
* convert obs code type string to obs code
* args   : char   *str   I      obs code string ("1C","1P","1Y",...)
* return : obs code (CODE_???)
* notes  : obs codes are based on RINEX 3.04
*-----------------------------------------------------------------------------*/
unsigned char obs2code(const char *obs)
{
    unsigned char i;

    for (i=1;*obscodes[i];i++) {
        if (strcmp(obscodes[i],obs)) continue;
        return i;
    }
    return CODE_NONE;
}
char *code2obs(const char code)
{
    if (code<=CODE_NONE||MAXCODE<code) return "";
    return obscodes[code];
}

/* GPS obs code to frequency -------------------------------------------------*/
static int code2freq_GPS(unsigned char code, double *freq)
{
    char *obs=code2obs(code);

    switch (obs[0]) {
        case '1': *freq=FREQL1; return 0; /* L1 */
        case '2': *freq=FREQL2; return 1; /* L2 */
        case '5': *freq=FREQL5; return 2; /* L5 */
    }
    return -1;
}
/* GLONASS obs code to frequency ---------------------------------------------*/
static int code2freq_GLO(unsigned char code, int fcn, double *freq)
{
    char *obs=code2obs(code);

    if (fcn<-7||fcn>6) return -1;

    switch (obs[0]) {
        case '1': *freq=FREQ1_GLO+DFRQ1_GLO*fcn; return 0; /* G1 */
        case '2': *freq=FREQ2_GLO+DFRQ2_GLO*fcn; return 1; /* G2 */
        case '3': *freq=FREQ3_GLO;               return 2; /* G3 */
        case '4': *freq=FREQ1a_GLO;              return 0; /* G1a */
        case '6': *freq=FREQ2a_GLO;              return 1; /* G2a */
    }
    return -1;
}
/* Galileo obs code to frequency ---------------------------------------------*/
static int code2freq_GAL(unsigned char code, double *freq)
{
    char *obs=code2obs(code);

    switch (obs[0]) {
        case '1': *freq=FREQL1; return 0; /* E1 */
        case '7': *freq=FREQE5b; return 1; /* E5b */
        case '5': *freq=FREQL5; return 2; /* E5a */
        case '6': *freq=FREQL6; return 3; /* E6 */
        case '8': *freq=FREQE5ab; return 4; /* E5ab */
    }
    return -1;
}
/* QZSS obs code to frequency ------------------------------------------------*/
static int code2freq_QZS(unsigned char code, double *freq)
{
    char *obs=code2obs(code);

    switch (obs[0]) {
        case '1': *freq=FREQL1; return 0; /* L1 */
        case '2': *freq=FREQL2; return 1; /* L2 */
        case '5': *freq=FREQL5; return 2; /* L5 */
        case '6': *freq=FREQL6; return 3; /* L6 */
    }
    return -1;
}
/* SBAS obs code to frequency ------------------------------------------------*/
static int code2freq_SBS(unsigned char code, double *freq)
{
    char *obs=code2obs(code);

    switch (obs[0]) {
        case '1': *freq=FREQL1; return 0; /* L1 */
        case '5': *freq=FREQL5; return 1; /* L5 */
    }
    return -1;
}
/* BDS obs code to frequency -------------------------------------------------*/
static int code2freq_BDS(unsigned char code, double *freq)
{
    char *obs=code2obs(code);

    switch (obs[0]) {
        case '1': *freq=FREQL1;    return 0; /* B1C */
        case '2': *freq=FREQ1_CMP; return 0; /* B1I */
        case '7': *freq=FREQ2_CMP; return 1; /* B2b */
        case '5': *freq=FREQL5;    return 2; /* B2a */
        case '6': *freq=FREQ3_CMP; return 3; /* B3 */
        case '8': *freq=FREQE5ab;  return 4; /* B2ab */
    }
    return -1;
}
/* NavIC obs code to frequency -----------------------------------------------*/
static int code2freq_IRN(unsigned char code, double *freq)
{
    char *obs=code2obs(code);

    switch (obs[0]) {
        case '5': *freq=FREQL5; return 0; /* L5 */
        case '9': *freq=FREQs; return 1; /* S */
    }
    return -1;
}
/* system and obs code to frequency index --------------------------------------
* convert system and obs code to frequency index
* args   : int    sys       I   satellite system (SYS_???)
*          unsigned char code     I   obs code (CODE_???)
* return : frequency index (-1: error)
*                       0     1     2     3     4
*           --------------------------------------
*            GPS       L1    L2    L5     -     -
*            GLONASS   G1    G2    G3     -     -  (G1=G1,G1a,G2=G2,G2a)
*            Galileo   E1    E5b   E5a   E6   E5ab
*            QZSS      L1    L2    L5    L6     -
*            SBAS      L1     -    L5     -     -
*            BDS       B1    B2b   B2a   B3   B2ab
*            NavIC     L5     S     -     -     -
*-----------------------------------------------------------------------------*/
int code2idx(int sys, unsigned char code)
{
    double freq;

    switch (sys) {
        case SYS_GPS: return code2freq_GPS(code,&freq);
        case SYS_GLO: return code2freq_GLO(code,0,&freq);
        case SYS_GAL: return code2freq_GAL(code,&freq);
        case SYS_QZS: return code2freq_QZS(code,&freq);
        case SYS_SBS: return code2freq_SBS(code,&freq);
        case SYS_CMP: return code2freq_BDS(code,&freq);
        case SYS_IRN: return code2freq_IRN(code,&freq);
    }
    return -1;
}
/* get code priority -----------------------------------------------------------
* get code priority for multiple codes in a frequency
* args   : int    sys       I   system (SYS_???)
*          uint8_t code     I   obs code (CODE_???)
*          char   *opt      I   code options (NULL:no option)
* return : priority (15:highest-1:lowest,0:error)
*-----------------------------------------------------------------------------*/
int getcodepri(int sys, unsigned char code, const char *opt)
{
    const char *p,*optstr;
    char *obs,str[8]="";
    int i,j;

    switch (sys) {
        case SYS_GPS: i=0; optstr="-GL%2s"; break;
        case SYS_GLO: i=1; optstr="-RL%2s"; break;
        case SYS_GAL: i=2; optstr="-EL%2s"; break;
        case SYS_QZS: i=3; optstr="-JL%2s"; break;
        case SYS_SBS: i=4; optstr="-SL%2s"; break;
        case SYS_CMP: i=5; optstr="-CL%2s"; break;
        case SYS_IRN: i=6; optstr="-IL%2s"; break;
        default: return 0;
    }
    if ((j=code2idx(sys,code))<0) return 0;
    obs=code2obs(code);

    /* parse code options */
    for (p=opt;p&&(p=strchr(p,'-'));p++) {
        if (sscanf(p,optstr,str)<1||str[0]!=obs[0]) continue;
        return str[1]==obs[1]?15:0;
    }
    /* search code priority */
    return (p=strchr(codepris[i][j],obs[1]))?14-(int)(p-codepris[i][j]):0;
}
/* set signal index ----------------------------------------------------------*/
static void set_index(double ver, int sys, const char *opt,
                      char tobs[MAXOBSTYPE][4], sigind_t *ind)
{
    const char *p;
    char str[8],*optstr="";
    double shift;
    int i,j,k,n;

    for (i=n=0;*tobs[i];i++,n++) {
        ind->code[i]=obs2code(tobs[i]+1);
        ind->type[i]=(p=strchr(obstyps,tobs[i][0]))?(int)(p-obstyps):0;
        ind->frqIdx[i]=code2idx(sys,ind->code[i]);
        ind->pri[i]=getcodepri(sys,ind->code[i],opt);
        ind->pos[i]=-1;
    }
    /* parse phase shift options */
    switch (sys) {
        case SYS_GPS: optstr="-GL%2s=%lf"; break;
        case SYS_GLO: optstr="-RL%2s=%lf"; break;
        case SYS_GAL: optstr="-EL%2s=%lf"; break;
        case SYS_QZS: optstr="-JL%2s=%lf"; break;
        case SYS_SBS: optstr="-SL%2s=%lf"; break;
        case SYS_CMP: optstr="-CL%2s=%lf"; break;
        case SYS_IRN: optstr="-IL%2s=%lf"; break;
    }
    for (p=opt;p&&(p=strchr(p,'-'));p++) {
        if (sscanf(p,optstr,str,&shift)<2) continue;
        for (i=0;i<n;i++) {
            if (strcmp(code2obs(ind->code[i]),str)) continue;
            ind->shift[i]=shift;
            printf("phase shift: sys=%2d tobs=%s shift=%.3f\n",sys,
                  tobs[i],shift);
        }
    }
    /* assign index for highest priority code */
    for (i=0;i<NFREQ;i++) {
        for (j=0,k=-1;j<n;j++) {
            if (ind->frqIdx[j]==i&&ind->pri[j]&&(k<0||ind->pri[j]>ind->pri[k])) {
                k=j;
            }
        }
        if (k<0) continue;

        for (j=0;j<n;j++) {
            if (ind->code[j]==ind->code[k]) ind->pos[j]=i;
        }
    }
    /* assign index of extended observation data */
    for (i=0;i<NEXOBS;i++) {
        for (j=0;j<n;j++) {
            if (ind->code[j]&&ind->pri[j]&&ind->pos[j]<0) break;
        }
        if (j>=n) break;

        for (k=0;k<n;k++) {
            if (ind->code[k]==ind->code[j]) ind->pos[k]=NFREQ+i;
        }
    }
    /* list rejected observation types */
    for (i=0;i<n;i++) {
        if (!ind->code[i]||!ind->pri[i]||ind->pos[i]>=0) continue;
        printf("reject obs type: sys=%2d, obs=%s\n",sys,tobs[i]);
    }
    ind->n=n;

}

bool readobsnav(gtime_t ts, gtime_t te, double ti,const char** infile,const int* index,int n,
                const prcopt_t* prcopt,obs_t* obs,nav_t* nav,sta_t* sta)
{
    int i,j,ind=0,nobs=0,rcv=0,info;
    sta_t statmp;
    obs->data = NULL,obs->n = 0,obs->nmax = 0;
    nav->eph = NULL , nav->n = 0, nav->nmax = 0;
    nav->geph = NULL, nav->ng = 0, nav->ngmax = 0;
    for(i=0;i<n;i++){
        if(index[i]!=ind){
            if(obs->n > nobs) rcv=rcv+1;
            ind=index[i]; 
            nobs=obs->n;
        }
        // read rinex obs and nav file 
        if(rcv<=2){
            info = readrnxt(infile[i],rcv,ts,te,ti,prcopt->rnxopt[rcv<=1?1:2],obs,nav,&sta[rcv]);
            if(info<=0){
                 printf("File reading error : %s\n",infile[i]);
            }
        }else{
            info = readrnxt(infile[i],rcv,ts,te,ti,prcopt->rnxopt[rcv<=1?1:2],obs,nav,&statmp);
            if(info<=0){
                 printf("File reading error : %s\n",infile[i]);
            }
        }
        if (info<0){
            return false;
        }
    }
    if(obs->n<=0)
    {
        printf("Error : no observation data!\n");
        return false;
    }
    if(nav->n <= 0 && nav->ng <= 0 && nav->ns <= 0)
    {
        printf("Error : no navigation data!\n");
        return false;
    }
    
    sortobs(obs);
    uniqnav(nav);
    return true;
}

int readrnxt(const char *filepath,int rcv,gtime_t ts,gtime_t te,double ti,const char *opt,
            obs_t* obs,nav_t* nav,sta_t* sta)
{
    int status = 0;
    char type = ' ';
    status = readrnxfile(filepath,ts,te,ti,opt,0,rcv,&type,obs,nav,sta);  // -1,0: error, 1-normal

    return status;
}

int readrnxfile(const char *filepath,gtime_t ts,gtime_t te,double ti,const char* opt, 
                int flag,int index,char* type,obs_t* obs,nav_t* nav,sta_t* sta)
{
    int status, info;
    init_sta(sta);
    FILE *fp;
    if(!(fp = fopen(filepath, "r"))){
        printf("File opening error : %s\n",filepath);
        status = -1;
        return status;
    }
    status = readrnxfp(fp,ts,te,ti,opt,flag,index,type,obs,nav,sta);
    close(fp);
    return status;
}

int readrnxfp(FILE* fp,gtime_t ts,gtime_t te,double ti,const char* opt, 
                    int flag,int index,char* type,obs_t* obs,nav_t* nav,sta_t* sta)
{
    double ver;
    int sys,tsys,status;
    char tobs[NUMSYS][MAXOBSTYPE][4] = {{""}};
    status = readrnxh(fp,&ver,type,&sys,&tsys,tobs,obs,nav,sta);
    if(status == 0)
    {
        printf("Error reading rinex header\n");
        return status;
    }
    switch (*type)
    {
        case 'O':
            status = readrnxobs(fp,ts,te,ti,opt,index,ver,&tsys,tobs,obs,sta);
            break;
        case 'N':
            status = readrnxnav(fp, opt, ver, sys, nav);
            break;
        case 'G':
            status = readrnxnav(fp, opt, ver, SYS_GLO, nav);
            break;
        case 'H':
            status = readrnxnav(fp, opt, ver, SYS_SBS, nav);
            break;
        case 'J':
            status = readrnxnav(fp, opt, ver, SYS_QZS, nav); /* extension */
            break;
        case 'L':
            status = readrnxnav(fp, opt, ver, SYS_GAL, nav); /* extension */
            break;
        default:
            break;
    }
    return status;
}

int readrnxh(FILE *fp, double *ver, char *type, int *sys, int *tsys,
                    char tobs[][MAXOBSTYPE][4],obs_t* obs,nav_t *nav, sta_t *sta)
{
    double bias;
    char buff[MAXRNXLEN];
    char *label;
    int i=0, block=0, sat=0, status=0;

    label = buff + 60;
    *ver = 2.10, *type = ' ', *sys=SYS_GPS, *tsys=TSYS_GPS;
    
    while (fgets(buff, MAXRNXLEN, fp))
    {
        if (strlen(buff) <= 60)
        {
            continue;
        }
        else if (strstr(label, "RINEX VERSION / TYPE"))
        {
            *ver = str2num(buff, 0, 9);
            
            *type = *(buff + 20);
            /* satellite system */
            switch (*(buff + 40))
            {
            case ' ':
            case 'G':
                *sys = SYS_GPS,*tsys = TSYS_GPS;
                break;
            case 'R':
                *sys = SYS_GLO,*tsys = TSYS_UTC;
                break;
            case 'E':
                *sys = SYS_GAL,*tsys = TSYS_GAL;
                break; /* v.2.12 */
            case 'S':
                *sys = SYS_SBS, *tsys = TSYS_GPS;
                break;
            case 'J':
                *sys = SYS_QZS,*tsys = TSYS_QZS;
                break; /* v.3.02 */
            case 'C':
                *sys = SYS_CMP,*tsys = TSYS_CMP;
                break; /* v.2.12 */
            case 'I':
                *sys = SYS_IRN,*tsys = TSYS_IRN;
                break; /* v.3.03 */
            case 'M':
                *sys = SYS_NONE,*tsys = TSYS_GPS;
                break; /* mixed */
            default:
                printf("not supported satellite system: %c\n", *(buff + 40));
                break;
            }
            continue;
        }
        else if (strstr(label, "PGM / RUN BY / DATE"))
        {
            continue;
        }
        else if (strstr(label, "COMMENT"))
        {
            continue;
        }
        switch (*type)
        {
        case 'O':
            decode_obsh(fp,buff,*ver,tsys,tobs,obs,nav,sta);
            break;
        case 'N':
            decode_navh(buff,nav);
            break;
        case 'G':
            decode_gnavh(buff,nav);
            break;    
        default:
            break;
        }
        if (strstr(label, "END OF HEADER")){
            status = 1;
            return status;
        }
        i = i+1;    

        if (i >= MAXPOSHEAD && *type == ' '){
            printf("no RINEX file\n");
            break; /* no RINEX file */
        }
            
    }

    return status;
}

int readrnxobs(FILE *fp, gtime_t ts, gtime_t te, double ti,
               const char *opt, int rcv, double ver, int *tsys,
            char tobs[][MAXOBSTYPE][4], obs_t *obs, sta_t *sta)
{
    obsd_t* data;
    int i,n,flag=0,status=0,slips[MAXSAT][NFREQ]={0};
    
    if(rcv>MAXRCV)
    {
        return status;
    }

    if (!(data = (obsd_t *)malloc(sizeof(obsd_t) * MAXOBS)))
    {
        printf("malloc error\n");
        return status;
    }

    while (true)
    {
        n = readrnxobsb(fp,opt,ver,tsys,tobs,&flag,data,sta);
        if(!(n>=0 && status>=0)) break;
        for (i = 0; i < n; i++)
        {
            /* UTC -> GPST */
            if (*tsys == TSYS_UTC)
            {
                data[i].time = utc2gpst(data[i].time);
            }
            /* save cycle slip */
            saveslips(slips, data + i);
        }
        /* screen data by time */
        if (n > 0 && !screent(data[0].time, ts, te, ti)){
            continue;
        }

        for (i = 0; i < n; i++)
        {
            /* restore cycle slip */
            restoreslips(slips, data + i);
            data[i].rcv = (unsigned char)rcv;
            /* save obs data */
            if ((status = addobsdata(obs, data + i)) < 0) break;
        } 
    }

    free(data);
    return status;
}

int readrnxobsb(FILE *fp, const char *opt, double ver, int *tsys,
                       char tobs[][MAXOBSTYPE][4], int *flag, obsd_t *data,
                       sta_t *sta)
{
    gtime_t time = {0};
    sigind_t  index[NUMSYS] = {{0}};
    obs_t obstmp;
    nav_t navtmp;
    char buff[MAXRNXLEN];
    int i=0,n=0,nsat=0,sats[MAXOBS]={0},mask,status=-1;
    mask = set_sysmask(opt);
    set_index(ver,SYS_GPS,opt,tobs[0],&index[0]);
    set_index(ver,SYS_GLO,opt,tobs[1],&index[1]);
    set_index(ver,SYS_GAL,opt,tobs[2],&index[2]);
    set_index(ver,SYS_QZS,opt,tobs[3],&index[3]);
    set_index(ver,SYS_SBS,opt,tobs[4],&index[4]);
    set_index(ver,SYS_CMP,opt,tobs[5],&index[5]);
    set_index(ver,SYS_IRN,opt,tobs[6],&index[6]);
    /* read record */
    while (fgets(buff,MAXRNXLEN,fp)) {

        /* decode observation epoch */
        if (i==0) {
            if ((nsat=decode_obsepoch(fp,buff,ver,&time,flag,sats))<=0 && (*flag != 5)) {
                continue;
            }
            // if (*flag == 5) {
            //     data[0].eventime = time;
            //     return 0;
            // }
        }
        else if ((*flag<=2||*flag==6)&&n<MAXOBS) {
            data[n].time=time;
            data[n].sat=(unsigned char)sats[i-1];

            /* decode RINEX observation data */
            if (decode_obsdata(fp,buff,ver,mask,index,data+n)) n++;
        }
        else if (*flag==3||*flag==4) { /* new site or header info follows */

            /* decode RINEX observation data file header */
            decode_obsh(fp,buff,ver,tsys,tobs,NULL,NULL,sta);
        }
        if (++i>nsat) return n;
    }

    return status;
}

void saveslips(unsigned char slips[][NFREQ + NEXOBS], obsd_t *data)
{
    int i;
    for (i = 0; i < NFREQ + NEXOBS; i++)
    {
        if (data->LLI[i] & 1)
            slips[data->sat - 1][i] |= LLI_SLIP;
    }
}

int screent(gtime_t time, gtime_t ts, gtime_t te, double tint)
{
    return (tint <= 0.0 || fmod(time2gpst(time, NULL) + DTTOL, tint) <= DTTOL * 2.0) &&
           (ts.time == 0 || timediff(time, ts) >= -DTTOL) &&
           (te.time == 0 || timediff(time, te) < DTTOL);
}

void restoreslips(unsigned char slips[][NFREQ + NEXOBS], obsd_t *data)
{
    int i;
    for (i = 0; i < NFREQ + NEXOBS; i++)
    {
        if (slips[data->sat - 1][i] & 1)
        {
            data->LLI[i] |= LLI_SLIP;
        }       
        slips[data->sat - 1][i] = 0;
    }
}

int addobsdata(obs_t *obs, const obsd_t *data)
{
    obsd_t *obs_data;

    if (obs->nmax <= obs->n)
    {
        if (obs->nmax <= 0)
            obs->nmax = NINCOBS;
        else
            obs->nmax *= 2;
        if (!(obs_data = (obsd_t *)realloc(obs->data, sizeof(obsd_t) * obs->nmax)))
        {
            printf("addobsdata: malloc error n=%dx%d\n", sizeof(obsd_t), obs->nmax);
            free(obs->data);
            obs->data = NULL;
            obs->n = obs->nmax = 0;
            return -1;
        }
        obs->data = obs_data;
    }
    obs->data[obs->n++] = *data;
    return 1;
}

/* read RINEX navigation data ------------------------------------------------*/
int readrnxnav(FILE *fp, const char *opt, double ver, int sys,nav_t *nav)
{
    eph_t eph;
    geph_t geph;
    seph_t seph;
    int status, type;

    printf("readrnxnav: ver=%.2f sys=%d\n", ver, sys);

    if (!nav) return 0;

    /* read RINEX navigation data body */
    while ((status = readrnxnavb(fp, opt, ver, sys, &type, &eph, &geph, &seph)) >= 0)
    {
        /* add ephemeris to navigation data */
        if (status)
        {
            switch (type)
            {
            case 1:
                status = add_geph(nav, &geph);
                break;
            case 2:
                status = add_seph(nav, &seph);
                break;
            default:
                status = add_eph(nav, &eph);
                break;
            }
            if (!status) return 0;
        }
    }
    return nav->n > 0 || nav->ng > 0 || nav->ns > 0;
}

int readrnxnavb(FILE *fp, const char *opt, double ver, int sys,
                int *type, eph_t *eph, geph_t *geph, seph_t *seph)
{
    gtime_t toc;
    double data[64];
    int i = 0, j, prn, sat = 0, sp = 3, mask;
    char buff[MAXRNXLEN], id[8] = "", *p;

    printf("readrnxnavb: ver=%.2f sys=%d\n", ver, sys);

    /* set system mask */
    mask = set_sysmask(opt);

    while (fgets(buff, MAXRNXLEN, fp))
    {
        if (i == 0)
        {
            /* decode satellite field */
            if (ver >= 3.0 || sys == SYS_GAL || sys == SYS_QZS)
            { /* ver.3 or GAL/QZS */
                sprintf(id, "%.3s", buff);
                sat = satid2no(id);
                sp = 4;
                if (ver >= 3.0)
                {
                    sys = satsys(sat, NULL);
                    if (!sys)
                    {
                        sys = (id[0] == 'S') ? SYS_SBS : ((id[0] == 'R') ? SYS_GLO : SYS_GPS);
                    }
                }
            }
            else
            {
                prn = (int)str2num(buff, 0, 2);

                if (sys == SYS_SBS)
                {
                    sat = satno(SYS_SBS, prn + 100);
                }
                else if (sys == SYS_GLO)
                {
                    sat = satno(SYS_GLO, prn);
                }
                else if (93 <= prn && prn <= 97)
                { /* extension */
                    sat = satno(SYS_QZS, prn + 100);
                }
                else
                    sat = satno(SYS_GPS, prn);
            }
            /* decode Toc field */
            if (str2time(buff + sp, 0, 19, &toc))
            {
                printf("rinex nav toc error: %23.23s\n", buff);
                return 0;
            }
            /* decode data fields */
            for (j = 0, p = buff + sp + 19; j < 3; j++, p += 19)
            {
                data[i++] = str2num(p, 0, 19);
            }
        }
        else
        {
            /* decode data fields */
            for (j = 0, p = buff + sp; j < 4; j++, p += 19)
            {
                data[i++] = str2num(p, 0, 19);
            }
            /* decode ephemeris */
            if (sys == SYS_GLO && i >= 15)
            {
                if (!(mask & sys))
                    return 0;
                *type = 1;
                return decode_geph(ver, sat, toc, data, geph);
            }
            else if (sys == SYS_SBS && i >= 15)
            {
                if (!(mask & sys))
                    return 0;
                *type = 2;
                return decode_seph(ver, sat, toc, data, seph);
            }
            else if (i >= 31)
            {
                if (!(mask & sys))
                    return 0;
                *type = 0;
                return decode_eph(ver, sat, toc, data, eph);
            }
        }
    }
    return -1;
}

void freeobsnav(obs_t *obs, nav_t* nav)
{
    if(obs)
    {
        if(obs->data)
        {
            free(obs->data);
            obs->data = NULL;
        }
        obs->n = 0, obs->nmax = 0;
    }
    if(nav)
    {
        if(nav->eph)
        {
            free(nav->eph);
            nav->eph = NULL;
        }
        nav->n = 0,nav->nmax = 0;

        if(nav->geph)
        {
            free(nav->geph);
            nav->geph = NULL;
        }
        nav->ng = 0,nav->ngmax = 0;

        if(nav->seph)
        {
            free(nav->seph);
            nav->seph = NULL;
        }
        nav->ns = 0,nav->nsmax = 0;
    }
}